
/*
 * DTST - Special disk test - finds slow sectors
 *
 * Copyright 1988 Samuel H. Smith; All rights reserved.
 *
 * Written 13-feb-88 (rev. 21-may-88)
 *
 */

#define VERSION  "DiskTest v1.4, (C)1988 S.H.Smith"

#include <stdio.h>
#include <dos.h>
#include <bios.h>
#include <ctype.h>
#include <errno.h>

#define SECSIZ   512
#define NSECT    31
char secbuf[NSECT * SECSIZ];
char patbuf[NSECT * SECSIZ];

#define DIRPERSEC 16


struct bootrec {		/* found at logical sector 0 (absolute read) */
	char jmp[3];
	char oem[8];
	int SectSiz;		/* bytes per sector */
	char ClustSiz;		/* sectors per cluster */
	int ResSecs;		/* reserved sectors before first FAT */
	char FatCnt;		/* number of FATs */
	int RootSiz;		/* max root directory entries (32 bytes each) */
	unsigned TotSecs;	/* total number or sectors */
	char Media;		/* media descriptor */
	int FatSize;		/* sectors per fat */
	int TrkSecs;		/* sectors per track */
	int HeadCnt;		/* number of heads */
	int HidnSec;		/* hidden sectors */

	char filler[0x200 - 0x1e];
} bootrec;


#define MAXBAD 2000
unsigned badclusts[MAXBAD];
unsigned badcount = 0;

unsigned reserved_sec;
unsigned cyl_secs;

int ptrack = -1;
int maxtime = 999;
int cluster;
int track;

int slows = 0;
int new_bad = 0;
int write_check = 0;



/* --------------------------------------------------------------- */
unsigned sec2clust(unsigned sec)
{
	if (sec < reserved_sec)
		 return 0;
	else
		return 2 + (sec - reserved_sec) / bootrec.ClustSiz;
}


/* --------------------------------------------------------------- */
unsigned clust2sec(unsigned clust)
{
	 return (clust - 2) * bootrec.ClustSiz + reserved_sec;
}


/* --------------------------------------------------------------- */
void generate_pattern(void)
{
	int i;

	srand(biostime(0, 0L));

	for (i = 0; i < sizeof(patbuf); i++)
		patbuf[i] = rand();
}


/* --------------------------------------------------------------- */
int exclude_bad_blocks(unsigned sec, int num)
{
	int i;
	int cluster = sec2clust(sec);
	int track = sec / cyl_secs;

	if (num > bootrec.ClustSiz)
		return 0;

	for (i = 0; i < badcount; i++) {
		if (badclusts[i] == cluster) {
			printf("\r  Track %d, cluster %u - Already marked bad\n",
			       track, cluster);
			return 1;
		}
	}

	return 0;
}


/* --------------------------------------------------------------- */
void report_cluster_number(unsigned sec, int num, unsigned speed)
{
	if (sec < reserved_sec)
		 printf("\r  Track %d, sector %u, speed %u mS ",
			 track, sec, speed);
	else if (num > bootrec.ClustSiz)
		printf("\r  Track %d, clusters %u-%u, speed %u mS ",
		       track, cluster,
		       cluster + (num / bootrec.ClustSiz) - 1, speed);
	else
		printf("\r  Track %d, cluster %u, speed %u mS ",
		       track, cluster, speed);
}


/* --------------------------------------------------------------- */
void rewrite_block(char disk, unsigned sec, int num)
 /* perform a READ/WRITE/READ/ReWRITE test */
{
	int result;

	result = abswrite(disk - 'A', num, sec, patbuf);
	if (result)
		printf("- PATTERN WRITE FAILED!\n");

	result = absread(disk - 'A', num, sec, patbuf);
	if (result)
		printf("- PATTERN READ FAILED!\n");

	/* replace original data */
	result = abswrite(disk - 'A', num, sec, secbuf);
	if (result) {
		printf("- ORIGINAL RE-WRITE FAILED!\n\n");

		result = abswrite(disk - 'A', num, sec, secbuf);
		if (result) {
			printf("- SECOND RE-WRITE FAILED!\n\n");
			printf("PROGRAM ABORTED - POSSIBLE DATA LOSS!\n");
			exit(1);
		}
	}
}


/* --------------------------------------------------------------- */
unsigned check_block(char disk, unsigned sec, int num)
 /* test a sector/block, return elapsed time for read.  time < 0 on errors */
{
	int result;
	long start;
	int elapsed;
	unsigned speed;

	cluster = sec2clust(sec);
	track = sec / cyl_secs;

	if (exclude_bad_blocks(sec, num))
		return 0;

	start = biostime(0, 0L);
	result = absread(disk - 'A', num, sec, secbuf);
	elapsed = biostime(0, start) - start;
	speed = elapsed * 18;

	if ((elapsed > maxtime) || (result) || (track != ptrack)) {
		ptrack = track;
		report_cluster_number(sec, num, speed);
	}


	/* extablish max time for read based on SECOND read request */
	if (maxtime == 999)
		maxtime = 888;
	else if (maxtime == 888)
		maxtime = (elapsed + 1) * 3;


	/* report errors or slow reads */
	if (result) {
		printf("- HARD ERROR!\n");
		if (num <= bootrec.ClustSiz)
			new_bad++;
	}
	else if (elapsed > maxtime) {
		printf("- SLOW! (soft errors)\n");
		slows++;
		result = 999;
	}

	if ((result == 0) && write_check)
		rewrite_block(disk, sec, num);

	if (result)
		return -elapsed;
	else
		return elapsed;
}


/* --------------------------------------------------------------- */
void test_range(char disk, unsigned low, unsigned last)
{
	unsigned current;
	unsigned single;
	int elapsed;
	int nsect;

	if (write_check)
		printf("\nPerforming READ/WRITE test:\n");
	else
		printf("\nPerforming READ test:\n");

	nsect = bootrec.TrkSecs;
	if (nsect > NSECT)
		nsect = NSECT;	/* special case for large-track disks
				 * (bernouli) */

	for (current = low; current <= last - nsect + 1; current += nsect) {
		elapsed = check_block(disk, current, nsect);

		/*
		 * if there was an error then switch to single sectors and
		 * rescan the sectors in question 
		 */
		if ((elapsed < 0) && (nsect > 1)) {
			for (single = current;
			     single < (current + nsect);
			     single += bootrec.ClustSiz) {
				check_block(disk, single, 1);
			}
			printf("\r%78s\r", " ");
		}
	}

	printf("\n\nTest completed:\n  %d slow clusters\n  %d new bad clusters\n",
	       slows, new_bad);
}


/* --------------------------------------------------------------- */
void get_disk_information(char disk)
{
#define BADCLUST 0xFFF7
#define FATS_PER_FATSEC (SECSIZ / sizeof(unsigned))
	unsigned fatbuf[FATS_PER_FATSEC];
	unsigned fatsec;
	unsigned bufpos;
	unsigned cluster;
	int result;

        printf("\nDisk %c:\n", disk);

	result = absread(disk - 'A', 1, 0, &bootrec);
	if (result) {
		printf("CANNOT READ BOOT RECORD\n");
		exit(1);
	}

	reserved_sec = bootrec.ResSecs +
		(bootrec.FatCnt * bootrec.FatSize) +
		(bootrec.RootSiz / DIRPERSEC);

	cyl_secs = bootrec.TrkSecs * bootrec.HeadCnt;

	printf("  Number of heads. . . . %d\n", bootrec.HeadCnt);
	printf("  Number of tracks . . . %d\n", bootrec.TotSecs / cyl_secs);
	printf("  Sectors per track. . . %d\n", bootrec.TrkSecs);
	printf("  Sectors per cluster. . %d\n", bootrec.ClustSiz);
	printf("  Total clusters . . . . %u\n", sec2clust(bootrec.TotSecs));

	printf("  Known bad clusters . . ");

	for (cluster = 0, fatsec = 1; fatsec <= bootrec.FatSize; fatsec++) {
		result = absread(disk - 'A', 1, fatsec, fatbuf);
		if (result) {
			printf("\n\nERROR READING FAT SECTOR %u\n",
			       fatsec);
			exit(1);
		}

		for (bufpos = 0; bufpos < FATS_PER_FATSEC; bufpos++, cluster++) {
			if (fatbuf[bufpos] == BADCLUST) {
				badclusts[badcount] = cluster;
				badcount++;

				printf("%-5u ", cluster);
				if ((badcount % 9) == 0)
					printf("\n                         ");
			}
		}
	}

	if (badcount == 0)
		printf("None");
	printf("\n");
}


/* --------------------------------------------------------------- */
void usage(void)
{
        printf("\n");
	printf("Usage:  disktest d: [<lowclust> <hiclust>] [-W]\n");
	printf("-W enables the nondestructive READ/WRITE test.\n\n");
	printf("DTST is a small utility that is used to check out your hard disks.  It\n");
        printf("has the special ability to report 'slow' sectors.  These are sectors\n");
	printf("that read properly, but often require retries.  I feel that such sectors\n");
	printf("are on the verge of failure and should quickly be locked out before they\n");
	printf("cause any harm.  USE AT YOUR OWN RISK.\n");
	exit(1);
}


/* --------------------------------------------------------------- */
void decode_option(char *option)
{
        if (strcmp(option, "-W") == 0)
		 write_check = 1;
	else
		usage();
}


/* --------------------------------------------------------------- */
main(int argc, char **argv)
{
	char disk;
	unsigned lowsec, hisec;

        printf("\n%s\n", VERSION);

	if (argc == 1)
		usage();

	disk = argv[1][0];
	if (islower(disk))
		disk = toupper(disk);

	get_disk_information(disk);
	lowsec = 0;
	hisec = bootrec.TotSecs;

	switch (argc) {
	case 2:
		break;		/* default to whole disk */

	case 3:
		decode_option(argv[2]);
		break;

	case 5:
		decode_option(argv[4]);

	case 4:
		lowsec = clust2sec(atoi(argv[2]));
		hisec = clust2sec(atoi(argv[3]));
		break;

	default:
		usage();
	}

	generate_pattern();
	test_range(disk, lowsec, hisec);

        exit(new_bad);
}
